      SUBROUTINE PCTRTRI( UPLO, DIAG, N, A, IA, JA, DESCA, INFO )
*
*  -- ScaLAPACK routine (version 1.0) --
*     University of Tennessee, Knoxville, Oak Ridge National Laboratory,
*     and University of California, Berkeley.
*     February 28, 1995
*
*     .. Scalar Arguments ..
      CHARACTER          DIAG, UPLO
      INTEGER            IA, INFO, JA, N
*     ..
*     .. Array Arguments ..
      INTEGER            DESCA( * )
      COMPLEX            A( * )
*     ..
*
*  Purpose
*  =======
*
*  PCTRTRI computes the inverse of a upper or lower triangular
*  distributed matrix sub( A ) = A(IA:IA+N-1,JA:JA+N-1).
*
*  Notes
*  =====
*
*  A description vector is associated with each 2D block-cyclicly dis-
*  tributed matrix.  This vector stores the information required to
*  establish the mapping between a matrix entry and its corresponding
*  process and memory location.
*
*  In the following comments, the character _ should be read as
*  "of the distributed matrix".  Let A be a generic term for any 2D
*  block cyclicly distributed matrix.  Its description vector is DESCA:
*
*  NOTATION        STORED IN  EXPLANATION
*  --------------- ---------- ------------------------------------------
*  M_A    (global) DESCA( 1 ) The number of rows in the distributed
*                             matrix.
*  N_A    (global) DESCA( 2 ) The number of columns in the distributed
*                             matrix.
*  MB_A   (global) DESCA( 3 ) The blocking factor used to distribute
*                             the rows of the matrix.
*  NB_A   (global) DESCA( 4 ) The blocking factor used to distribute
*                             the columns of the matrix.
*  RSRC_A (global) DESCA( 5 ) The process row over which the first row
*                             of the matrix is distributed.
*  CSRC_A (global) DESCA( 6 ) The process column over which the first
*                             column of the matrix is distributed.
*  CTXT_A (global) DESCA( 7 ) The BLACS context handle, indicating the
*                             BLACS process grid A is distributed over.
*                             The context itself is global, but the handle
*                             (the integer value) may vary.
*  LLD_A  (local)  DESCA( 8 ) The leading dimension of the local array
*                             storing the local blocks of the distri-
*                             buted matrix A. LLD_A >= MAX(1,LOCp(M_A)).
*
*  Let K be the number of rows or columns of a distributed matrix,
*  and assume that its process grid has dimension p x q.
*  LOCp( K ) denotes the number of elements of K that a process
*  would receive if K were distributed over the p processes of its
*  process column.
*  Similarly, LOCq( K ) denotes the number of elements of K that a
*  process would receive if K were distributed over the q processes of
*  its process row.
*  The values of LOCp() and LOCq() may be determined via a call to the
*  ScaLAPACK tool function, NUMROC:
*          LOCp( M ) = NUMROC( M, MB_A, MYROW, RSRC_A, NPROW ),
*          LOCq( N ) = NUMROC( N, NB_A, MYCOL, CSRC_A, NPCOL ).
*
*  Arguments
*  =========
*
*  UPLO    (global input) CHARACTER
*          Specifies whether the distributed matrix sub( A ) is upper
*          or lower triangular:
*          = 'U':  Upper triangular,
*          = 'L':  Lower triangular.
*
*  DIAG    (global input) CHARACTER
*          Specifies whether or not the distributed matrix sub( A )
*          is unit triangular:
*          = 'N':  Non-unit triangular,
*          = 'U':  Unit triangular.
*
*  N       (global input) INTEGER
*          The number of rows and columns to be operated on, i.e. the
*          order of the distributed submatrix sub( A ). N >= 0.
*
*  A       (local input/local output) COMPLEX pointer into the
*          local memory to an array of dimension (LLD_A,LOCq(JA+N-1)),
*          this array contains the local pieces of the triangular matrix
*          sub( A ).  If UPLO = 'U', the leading N-by-N upper triangular
*          part of the matrix sub( A ) contains the upper triangular
*          matrix to be inverted, and the strictly lower triangular
*          part of sub( A ) is not referenced.
*          If UPLO = 'L', the leading N-by-N lower triangular part of
*          the matrix sub( A ) contains the lower triangular matrix,
*          and the strictly upper triangular part of sub( A ) is not
*          referenced.
*
*  IA      (global input) INTEGER
*          A's global row index, which points to the beginning of the
*          submatrix which is to be operated on.
*
*  JA      (global input) INTEGER
*          A's global column index, which points to the beginning of
*          the submatrix which is to be operated on.
*
*  DESCA   (global and local input) INTEGER array of dimension 8
*          The array descriptor for the distributed matrix A.
*
*  INFO    (global output) INTEGER
*          = 0:  successful exit
*          < 0:  If the i-th argument is an array and the j-entry had
*                an illegal value, then INFO = -(i*100+j), if the i-th
*                argument is a scalar and had an illegal value, then
*                INFO = -i.
*
*  =====================================================================
*            > 0:  If INFO = K, A(IA+K-1,JA+K-1) is exactly zero.  The
*                  triangular matrix sub( A ) is singular and its
*                  inverse can not be computed.
*
*  ====================================================================
*
*     .. Parameters ..
      COMPLEX            ZERO, ONE
      PARAMETER          ( ZERO = 0.0E+0, ONE = 1.0E+0 )
*     ..
*     .. Local Scalars ..
      LOGICAL            NOUNIT, UPPER
      INTEGER            I, ICOFF, ICTXT, IROFF, ICURCOL, ICURROW,
     $                   IDUMMY, II, IOFFA, J, JB, JJ, JN, LDA, MYCOL,
     $                   MYROW, NN, NPCOL, NPROW
*     ..
*     .. Local Arrays ..
      INTEGER            IDUM1( 2 ), IDUM2( 2 )
*     ..
*     .. External Subroutines ..
      EXTERNAL           BLACS_GRIDINFO, CHK1MAT, IGAMX2D, INFOG2L,
     $                   PCHK1MAT, PCTRTI2, PCTRMM, PCTRSM,
     $                   PXERBLA
*     ..
*     .. External Functions ..
      LOGICAL            LSAME
      INTEGER            ICEIL
      EXTERNAL           ICEIL, LSAME
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          ICHAR, MIN, MOD
*     ..
*     .. Executable Statements ..
*
*     Get grid parameters
*
      ICTXT = DESCA( 7 )
      CALL BLACS_GRIDINFO( ICTXT, NPROW, NPCOL, MYROW, MYCOL )
*
*     Test input parameters
*
      INFO = 0
      IF( NPROW.EQ.-1 ) THEN
         INFO = -707
      ELSE
         UPPER = LSAME( UPLO, 'U' )
         NOUNIT = LSAME( DIAG, 'N' )
*
         CALL CHK1MAT( N, 3, N, 3, IA, JA, DESCA, 7, INFO )
         IF( INFO.EQ.0 ) THEN
            IROFF = MOD( IA-1, DESCA( 3 ) )
            ICOFF = MOD( JA-1, DESCA( 4 ) )
            IF( .NOT.UPPER .AND. .NOT.LSAME( UPLO, 'L' ) ) THEN
               INFO = -1
            ELSE IF( .NOT.NOUNIT .AND. .NOT.LSAME( DIAG, 'U' ) ) THEN
               INFO = -2
            ELSE IF( IROFF.NE.ICOFF .OR. IROFF.NE.0 ) THEN
               INFO = -6
            ELSE IF( DESCA( 3 ).NE.DESCA( 4 ) ) THEN
               INFO = -704
            END IF
         END IF
*
         IF( UPPER ) THEN
            IDUM1( 1 ) = ICHAR( 'U' )
         ELSE
            IDUM1( 1 ) = ICHAR( 'L' )
         END IF
         IDUM2( 1 ) = 1
         IF( NOUNIT ) THEN
            IDUM1( 2 ) = ICHAR( 'N' )
         ELSE
            IDUM1( 2 ) = ICHAR( 'U' )
         END IF
         IDUM2( 2 ) = 2
*
         CALL PCHK1MAT( N, 3, N, 3, IA, JA, DESCA, 7, 2, IDUM1, IDUM2,
     $                  INFO )
      END IF
*
      IF( INFO.NE.0 ) THEN
         CALL PXERBLA( ICTXT, 'PCTRTRI', INFO )
         RETURN
      END IF
*
*     Quick return if possible
*
      IF( N.EQ.0 )
     $   RETURN
*
*     Check for singularity if non-unit.
*
      JN = MIN( ICEIL( JA, DESCA( 4 ) ) * DESCA( 4 ), JA+N-1 )
      IF( NOUNIT ) THEN
         CALL INFOG2L( IA, JA, DESCA, NPROW, NPCOL, MYROW, MYCOL,
     $                 II, JJ, ICURROW, ICURCOL )
*
*        Handle first block separately
*
         JB = JN-JA+1
         LDA = DESCA( 8 )
         IF( MYROW.EQ.ICURROW .AND. MYCOL.EQ.ICURCOL ) THEN
            IOFFA = II+(JJ-1)*LDA
            DO 10 I = 0, JB-1
               IF( A( IOFFA ).EQ.ZERO .AND. INFO.EQ.0 )
     $            INFO = I + 1
               IOFFA = IOFFA + LDA + 1
   10       CONTINUE
         END IF
         IF( MYROW.EQ.ICURROW )
     $      II = II + JB
         IF( MYCOL.EQ.ICURCOL )
     $      JJ = JJ + JB
         ICURROW = MOD( ICURROW+1, NPROW )
         ICURCOL = MOD( ICURCOL+1, NPCOL )
*
*        Loop over remaining blocks of columns
*
         DO 30 J = JN+1, JA+N-1, DESCA( 4 )
            JB = MIN( JA+N-J, DESCA( 4 ) )
            IF( MYROW.EQ.ICURROW .AND. MYCOL.EQ.ICURCOL ) THEN
               IOFFA = II+(JJ-1)*LDA
               DO 20 I = 0, JB-1
                  IF( A( IOFFA ).EQ.ZERO .AND. INFO.EQ.0 )
     $               INFO = J + I - JA + 1
                  IOFFA = IOFFA + LDA + 1
   20          CONTINUE
            END IF
            IF( MYROW.EQ.ICURROW )
     $         II = II + JB
            IF( MYCOL.EQ.ICURCOL )
     $         JJ = JJ + JB
            ICURROW = MOD( ICURROW+1, NPROW )
            ICURCOL = MOD( ICURCOL+1, NPCOL )
   30    CONTINUE
         CALL IGAMX2D( ICTXT, 'All', ' ', 1, 1, INFO, 1, IDUMMY,
     $                 IDUMMY, -1, -1, MYCOL )
         IF( INFO.NE.0 )
     $      RETURN
      END IF
*
*     Use blocked code
*
      IF( UPPER ) THEN
*
*        Compute inverse of upper triangular matrix
*
         JB = JN-JA+1
*
*        Handle first block of column separately
*
         CALL PCTRTI2( UPLO, DIAG, JB, A, IA, JA, DESCA, INFO )
*
*        Loop over remaining block of columns
*
         DO 40 J = JN+1, JA+N-1, DESCA( 4 )
            JB = MIN( DESCA( 4 ), JA+N-J )
            I = IA + J - JA
*
*           Compute rows 1:j-1 of current block column
*
            CALL PCTRMM( 'Left', UPLO, 'No transpose', DIAG, J-JA, JB,
     $                   ONE, A, IA, JA, DESCA, A, IA, J, DESCA )
            CALL PCTRSM( 'Right', UPLO, 'No transpose', DIAG, J-JA,
     $                   JB, -ONE, A, I, J, DESCA, A, IA, J, DESCA )
*
*           Compute inverse of current diagonal block
*
            CALL PCTRTI2( UPLO, DIAG, JB, A, I, J, DESCA, INFO )
*
   40    CONTINUE
*
      ELSE
*
*        Compute inverse of lower triangular matrix
*
         NN = ( ( JA+N-2 ) / DESCA( 4 ) )*DESCA( 4 ) + 1
         DO 50 J = NN, JN+1, -DESCA( 4 )
            JB = MIN( DESCA( 4 ), JA+N-J )
            I = IA + J - JA
            IF( J+JB.LE.JA+N-1 ) THEN
*
*              Compute rows j+jb:ja+n-1 of current block column
*
               CALL PCTRMM( 'Left', UPLO, 'No transpose', DIAG,
     $                      JA+N-J-JB, JB, ONE, A, I+JB, J+JB, DESCA,
     $                      A, I+JB, J, DESCA )
               CALL PCTRSM( 'Right', UPLO, 'No transpose', DIAG,
     $                      JA+N-J-JB, JB, -ONE, A, I, J, DESCA,
     $                      A, I+JB, J, DESCA )
            END IF
*
*           Compute inverse of current diagonal block
*
            CALL PCTRTI2( UPLO, DIAG, JB, A, I, J, DESCA, INFO )
*
   50    CONTINUE
*
*        Handle the last block of columns separately
*
         JB = JN-JA+1
         IF( JA+JB.LE.JA+N-1 ) THEN
*
*           Compute rows ja+jb:ja+n-1 of current block column
*
            CALL PCTRMM( 'Left', UPLO, 'No transpose', DIAG, N-JB, JB,
     $                   ONE, A, IA+JB, JA+JB, DESCA, A, IA+JB, JA,
     $                   DESCA )
            CALL PCTRSM( 'Right', UPLO, 'No transpose', DIAG, N-JB, JB,
     $                   -ONE, A, IA, JA, DESCA, A, IA+JB, JA, DESCA )
         END IF
*
*        Compute inverse of current diagonal block
*
         CALL PCTRTI2( UPLO, DIAG, JB, A, IA, JA, DESCA, INFO )
*
      END IF
*
      RETURN
*
*     End PCTRTRI
*
      END
